#!/usr/bin/env python

"""
Inspect source files, obtaining details of classes and attributes.

Copyright (C) 2007, 2008 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.

--------

The central classes in this module are the following:

  * Class
  * Function
  * Module
  * InspectedModule (derived from Module)

All of the above support the Naming interface either explicitly or through
general conformance, meaning that all can be asked to provide their 'full_name'
using the method of that name.

Additionally, all of the above also support a dictionary interface in order to
access names within their defined scopes. Specific methods also exist in order
to distinguish between certain kinds of attributes:

  * Class: (class|all_class|instance|all)_attributes
  * Function: parameters, locals, all_locals
  * Module: module_attributes

These specific methods are useful in certain situations.

The above classes also provide a 'node' attribute, indicating the AST node where
each such object is defined.
"""

from micropython.common import *
import compiler.ast
from compiler.visitor import ASTVisitor
try:
    set
except NameError:
    from sets import Set as set

class InspectError(ProcessingError): pass

class AtLeast:

    "A special representation for numbers of a given value or greater."

    def __init__(self, count):
        self.count = count

    def __eq__(self, other):
        return 0

    __lt__ = __le__ = __eq__

    def __ne__(self, other):
        return 1

    def __gt__(self, other):
        if isinstance(other, AtLeast):
            return 0
        else:
            return self.count > other

    def __ge__(self, other):
        if isinstance(other, AtLeast):
            return 0
        else:
            return self.count >= other

    def __iadd__(self, other):
        if isinstance(other, AtLeast):
            self.count += other.count
        else:
            self.count += other
        return self

    def __radd__(self, other):
        if isinstance(other, AtLeast):
            return AtLeast(self.count + other.count)
        else:
            return AtLeast(self.count + other)

    def __repr__(self):
        return "AtLeast(%r)" % self.count

# Mix-ins and abstract classes.

class NamespaceDict:

    "A mix-in providing dictionary methods."

    def __init__(self, global_namespace=None):
        self.namespace = {}
        self.globals = set()
        self.global_namespace = global_namespace
        self.attr_position = 0

    def __getitem__(self, name):
        return self.namespace[name]

    def get(self, name, default=None):
        return self.namespace.get(name, default)

    def __setitem__(self, name, value):
        self.set(name, value)

    def set(self, name, value, single_assignment=1):

        """
        A more powerful set operation, making 'name' refer to 'value' whilst
        indicating whether a 'single_assignment' (true by default) occurs in
        this operation (or whether the operation covers potentially many
        assignments in the lifetime of a program).
        """

        if name in self.globals:
            self.global_namespace.set(name, value, 0)
        else:
            attr = self._set(name, value)

            # NOTE: Insist on assignments with known values.

            if value is not None:
                attr.update(value, single_assignment)

    def set_module(self, name, value):

        """
        A specialised set operation, making 'name' refer to 'value' in the
        context of making a module reference available in association with
        'name' as part of the import of that module or a submodule of that
        module.
        """

        attr = self._set(name, value)
        if attr.assignments is None:
            attr.assignments = 1
        attr.assignment_values.add(value)

    def _set(self, name, value):

        "The underlying set operation associating 'name' with 'value'."

        if not self.namespace.has_key(name):
            self.namespace[name] = Attr(self.attr_position, self, name, value)
            self.attr_position += 1
        return self.namespace[name]

    def __delitem__(self, name):
        del self.namespace[name]

    def has_key(self, name):
        return self.namespace.has_key(name)

    def keys(self):
        return self.namespace.keys()

    def values(self):
        return self.namespace.values()

    def items(self):
        return self.namespace.items()

    def make_global(self, name):
        if not self.namespace.has_key(name):
            self.globals.add(name)
        else:
            raise InspectError(self.full_name(), self.node, "Name %r is both global and local in %r" % (name, self.full_name()))

    def get_assignments(self, name):
        if self.assignments.has_key(name):
            return max(self.assignments[name], len(self.assignment_values[name]))
        else:
            return None

    def to_list(self, d):
        l = [None] * len(d.keys())
        for attr in d.values():
            l[attr.position] = attr
        return l

class Naming:

    "A mix-in providing naming conveniences."

    def full_name(self):
        if self.name is not None:
            return self.parent_name + "." + self.name
        else:
            return self.parent_name

# Program data structures.

class Attr:

    "An attribute entry."

    def __init__(self, position, parent, name, value=None, assignments=None):
        self.position = position
        self.parent = parent
        self.name = name
        self.value = value

        # Number of assignments per name.

        self.assignments = assignments
        self.assignment_values = set()

    def update(self, value, single_assignment):

        """
        Update the attribute, adding the 'value' provided to the known values
        associated with the attribute, changing the number of assignments
        according to the 'single_assignment' status of the operation, where
        a true value indicates that only one assignment is associated with the
        update, and a false value indicates that potentially many assignments
        may be involved.
        """

        if self.assignments is None:
            if single_assignment:
                self.assignments = 1
            else:
                self.assignments = AtLeast(1)
        else:
            if single_assignment:
                self.assignments += 1
            else:
                self.assignments += AtLeast(1)
        self.assignment_values.add(value)

    def __repr__(self):
        return "Attr(%r, %r, %r, %r, %r)" % (self.position, self.parent, self.name, self.value, self.assignments)

class Const:

    "A constant object."

    def __init__(self, value):
        self.value = value

        # Image generation details.

        self.location = None

    def __repr__(self):
        if self.location is not None:
            return "Const(%r, location=%r)" % (self.value, self.location)
        else:
            return "Const(%r)" % self.value

    def __eq__(self, other):
        return self.value == other.value

    def __hash__(self):
        return hash(self.value)

class Class(NamespaceDict, Naming):

    "An inspected class."

    def __init__(self, name, parent_name, global_namespace=None, node=None):
        NamespaceDict.__init__(self, global_namespace)
        self.name = name
        self.parent_name = parent_name
        self.node = node

        # Superclasses and attributes.

        self.bases = []
        self.instattr = set()           # instance attributes
        self.instattr_relocated = set() # instance attributes which do not have
                                        # the same position as those of the same
                                        # name in some superclasses

        # Caches.

        self.all_instattr = None        # cache for instance_attributes
        self.all_instattr_names = None  # from all_instattr
        self.all_classattr = None       # cache for all_class_attributes
        self.all_classattr_names = None # from all_classattr
        self.allattr = None             # cache for all_attributes
        self.allattr_names = None       # from allattr

        # Image generation details.

        self.location = None
        self.code_location = None

    def __repr__(self):
        if self.location is not None:
            return "Class(%r, %r, location=%r)" % (self.name, self.parent_name, self.location)
        else:
            return "Class(%r, %r)" % (self.name, self.parent_name)

    def add_base(self, base):
        self.bases.append(base)

    def add_instance_attribute(self, name):
        self.instattr.add(name)

    "Return the attribute names provided by this class only."

    class_attribute_names = NamespaceDict.keys

    def class_attributes(self):

        "Return class attributes provided by this class only."

        return self

    def all_class_attribute_names(self):

        "Return the attribute names provided by classes in this hierarchy."

        if self.all_classattr_names is None:
            self.all_class_attributes()
        return self.all_classattr_names

    def all_class_attributes(self):

        "Return all class attributes, indicating the class which provides them."

        if self.all_classattr is None:
            self.all_classattr = {}

            reversed_bases = self.bases[:]
            reversed_bases.reverse()
            for cls in reversed_bases:
                self.all_classattr.update(cls.all_class_attributes())

            # Record attributes provided by this class, along with their
            # positions.

            self.all_classattr.update(self.class_attributes())

        return self.all_classattr

    def instance_attribute_names(self):

        "Return the instance attribute names provided by the class."

        if self.all_instattr_names is None:
            self.instance_attributes()
        return self.all_instattr_names

    def instance_attributes(self):

        "Return instance-only attributes for instances of this class."

        if self.all_instattr is None:
            self.all_instattr = {}
            instattr = {}

            reversed_bases = self.bases[:]
            reversed_bases.reverse()

            # For the bases in reverse order, acquire instance attribute
            # details.

            for cls in reversed_bases:
                for name, attr in cls.instance_attributes().items():
                    if not instattr.has_key(name):
                        instattr[name] = set()
                    instattr[name].add(attr.position)

            # Record instance attributes provided by this class and its bases,
            # along with their positions.

            for name in self.instattr:
                if not instattr.has_key(name):
                    instattr[name] = set([-1]) # position not yet defined

            # Cache the attributes by converting the positioned attributes into
            # a dictionary.

            if not instattr:
                self.all_instattr = {}
            else:
                self.all_instattr = dict(self._get_position_list(instattr))

            self.all_instattr_names = self.all_instattr.keys()

        return self.all_instattr

    def _get_position_list(self, instattr):

        """
        Return the instance attributes, 'instattr', as a list indicating the
        positions of the attributes in the final instance structure.
        """

        positions = instattr.items()
        instarray = [None] * len(positions)

        # Get the positions in ascending order of list size, with lists
        # of the same size ordered according to their smallest position
        # value.

        positions.sort(self._cmp_positions)
        #print self.name, positions

        # Get the names in position order.

        held = []

        for name, pos in positions:
            pos = list(pos)
            if pos[0] != -1 and instarray[pos[0]] is None:
                instarray[pos[0]] = name, Attr(pos[0], None, name)
            else:
                if pos[0] != -1 or len(pos) > 1:
                    self.instattr_relocated.add(name)
                held.append((name, pos))

        for i, attr in enumerate(instarray):
            if attr is None:
                name, pos = held.pop()
                instarray[i] = name, Attr(i, None, name)

        return instarray

    def _cmp_positions(self, a, b):

        "Compare name plus position list operands 'a' and 'b'."

        name_a, list_a = a
        name_b, list_b = b
        if len(list_a) < len(list_b):
            return -1
        elif len(list_a) > len(list_b):
            return 1
        else:
            return cmp(min(list_a), min(list_b))

    def all_attribute_names(self):

        """
        Return the names of all attributes provided by instances of this class.
        """

        self.allattr_names = self.allattr_names or self.all_attributes().keys()
        return self.allattr_names

    def all_attributes(self):

        """
        Return all attributes for an instance, indicating either the class which
        provides them or that the instance itself provides them.
        """

        if self.allattr is None:
            self.allattr = {}
            self.allattr.update(self.all_class_attributes())
            for name, attr in self.instance_attributes().items():
                if self.allattr.has_key(name):
                    print "Instance attribute %r in %r overrides class attribute." % (name, self)
                self.allattr[name] = attr
        return self.allattr

class Function(NamespaceDict, Naming):

    "An inspected function."

    def __init__(self, name, parent_name, argnames, has_star, has_dstar, global_namespace=None, node=None):
        NamespaceDict.__init__(self, global_namespace)
        self.name = name
        self.parent_name = parent_name
        self.argnames = argnames
        self.has_star = has_star
        self.has_dstar = has_dstar
        self.node = node

        # Caches.

        self.localnames = None # cache for locals

        # Add parameters to the namespace.

        self._add_parameters(argnames)

        # Image generation details.

        self.location = None
        self.code_location = None

    def _add_parameters(self, argnames):
        for name in argnames:
            if isinstance(name, tuple):
                self._add_parameters(name)
            else:
                self[name] = None

    def __repr__(self):
        if self.location is not None:
            return "Function(%r, %r, %r, %r, %r, location=%r)" % (
                self.name, self.parent_name, self.argnames, self.has_star, self.has_dstar, self.location
                )
        else:
            return "Function(%r, %r, %r, %r, %r)" % (
                self.name, self.parent_name, self.argnames, self.has_star, self.has_dstar
                )

    def make_global(self, name):
        if name not in self.argnames and not self.has_key(name):
            self.globals.add(name)
        else:
            raise InspectError(self.full_name(), self.node, "Name %r is global and local in %r" % (name, self.full_name()))

    def parameters(self):

        """
        Return a dictionary mapping parameter names to their position in the
        parameter list.
        """

        parameters = {}
        for i, name in enumerate(self.argnames):
            parameters[name] = i
        return parameters

    def all_locals(self):

        "Return a dictionary mapping names to local and parameter details."

        return self

    def locals(self):

        "Return a dictionary mapping names to local details."

        if self.localnames is None:
            self.localnames = {}
            self.localnames.update(self.all_locals())
            for name in self.argnames:
                del self.localnames[name]
        return self.localnames

class UnresolvedName(NamespaceDict, Naming):

    "A module, class or function which was mentioned but could not be imported."

    def __init__(self, name, parent_name, global_namespace=None):
        NamespaceDict.__init__(self, global_namespace)
        self.name = name
        self.parent_name = parent_name

    def all_class_attributes(self):
        return {}

    def instance_attributes(self):
        return {}

    def __repr__(self):
        return "UnresolvedName(%r, %r)" % (self.name, self.parent_name)

class Module(NamespaceDict):

    "An inspected module's core details."

    def __init__(self, name):
        NamespaceDict.__init__(self, self)
        self.name = name

        # Complete lists of classes and functions.

        self.all_objects = set()

        # Constant records.

        self.constant_values = {}
        self.constant_list = None # cache for constants

        # Image generation details.

        self.location = None
        self.code_location = None

        # Original location details.

        self.node = None

    def full_name(self):
        return self.name

    def __repr__(self):
        if self.location is not None:
            return "Module(%r, location=%r)" % (self.name, self.location)
        else:
            return "Module(%r)" % self.name

    # Attribute methods.

    "Return the module attribute names provided by the module."

    module_attribute_names = NamespaceDict.keys

    def module_attributes(self):

        "Return a dictionary mapping names to module attributes."

        return self

    def constants(self):

        "Return a list of constants."

        if self.constant_list is None:
            self.constant_list = list(self.constant_values.values())

        return self.constant_list

# Program visitors.

class InspectedModule(ASTVisitor, Module):

    """
    An inspected module, providing core details via the Module superclass, but
    capable of being used as an AST visitor.
    """

    def __init__(self, name, importer=None):
        ASTVisitor.__init__(self)
        Module.__init__(self, name)
        self.visitor = self

        self.importer = importer
        self.loaded = 0

        # Current expression state.

        self.expr = None

        # Namespace state.

        self.in_init = 0        # Find instance attributes in __init__ methods.
        self.in_loop = 0        # Note loop "membership", affecting assignments.
        self.namespaces = []
        self.module = None

    def parse(self, filename):

        "Parse the file having the given 'filename'."

        module = compiler.parseFile(filename)
        self.process(module)

    def process(self, module):

        "Process the given 'module'."

        self.node = self.module = module
        processed = self.dispatch(module)
        if self.has_key("__all__"):
            all = self["__all__"]
            if isinstance(all, compiler.ast.List):
                for n in all.nodes:
                    self[n.value] = self.importer.add_module(self.name + "." + n.value)
        return processed

    def vacuum(self):

        "Vacuum the module namespace, removing unloaded module references."

        for name, value in self.items():
            if isinstance(value, Module) and not value.loaded:
                del self[name]

            # Complain about globals not initialised at the module level.

            if isinstance(value, Global):
                print "Warning: global %r in module %r not initialised at the module level." % (name, self.name)

    # Namespace methods.

    def store(self, name, obj):

        "Record attribute or local 'name', storing 'obj'."

        if not self.namespaces:
            self.set(name, obj, not self.in_loop)
        else:
            self.namespaces[-1].set(name, obj, not self.in_loop)

        # Record all non-local objects.

        if not (self.namespaces and isinstance(self.namespaces[-1], Function)):
            self.all_objects.add(obj)

    def store_instance_attr(self, name):

        "Record instance attribute 'name' in the current class."

        if self.in_init:

            # Current namespace is the function.
            # Previous namespace is the class.

            self.namespaces[-2].add_instance_attribute(name)

    def get_parent(self):
        return (self.namespaces[-1:] or [self])[0]

    # Visitor methods.

    def default(self, node, *args):
        raise InspectError(self.full_name(), node, "Node class %r is not supported." % node.__class__)

    def dispatch(self, node, *args):
        return ASTVisitor.dispatch(self, node, *args)

    def NOP(self, node):
        for n in node.getChildNodes():
            self.dispatch(n)
        return None

    visitAdd = NOP

    visitAnd = NOP

    visitAssert = NOP

    def visitAssign(self, node):
        self.expr = self.dispatch(node.expr)
        for n in node.nodes:
            self.dispatch(n)
        return None

    def visitAssAttr(self, node):
        expr = self.dispatch(node.expr)
        if isinstance(expr, Attr) and expr.name == "self":
            self.store_instance_attr(node.attrname)
        return None

    def visitAssList(self, node):
        for n in node.nodes:
            self.dispatch(n)
        return None

    def visitAssName(self, node):
        if isinstance(self.expr, Attr):
            self.store(node.name, self.expr.value)
        else:
            self.store(node.name, self.expr)
        return None

    visitAssTuple = visitAssList

    visitAugAssign = NOP

    visitBackquote = NOP

    visitBitand = NOP

    visitBitor = NOP

    visitBitxor = NOP

    visitBreak = NOP

    visitCallFunc = NOP

    def visitClass(self, node):
        if self.namespaces:
            print "Class %r in %r is not global: ignored." % (node.name, self.namespaces[-1].full_name())
        else:
            cls = Class(node.name, self.get_parent().full_name(), self, node)

            # Visit the base class expressions, attempting to find concrete
            # definitions of classes.

            for base in node.bases:
                expr = self.dispatch(base)
                if isinstance(expr, Attr):
                    if expr.assignments != 1:
                        raise InspectError(self.full_name(), node,
                            "Base class %r for %r is not constant." % (base, cls.full_name()))
                    else:
                        cls.add_base(expr.value)
                else: # if expr is None:
                    raise InspectError(self.full_name(), node,
                        "Base class %r for %r is not found: it may be hidden in some way." % (base, cls.full_name()))

            # Make a back reference from the node for code generation.

            node.unit = cls

            # Make an entry for the class.

            self.store(node.name, cls)

            self.namespaces.append(cls)
            self.dispatch(node.code)
            self.namespaces.pop()

        return None

    visitCompare = NOP

    def visitConst(self, node):
        const = Const(node.value)
        self.constant_values[node.value] = const
        return const

    visitContinue = NOP

    visitDecorators = NOP

    visitDict = NOP

    visitDiscard = NOP

    visitDiv = NOP

    visitEllipsis = NOP

    visitExec = NOP

    visitExpression = NOP

    visitFloorDiv = NOP

    def visitFor(self, node):
        self.in_loop = 1
        self.NOP(node)
        self.in_loop = 0

    def visitFrom(self, node):
        if self.importer is None:
            raise InspectError(self.full_name(), node,
                "Please use the micropython.Importer class for code which uses the 'from' statement.")

        module = self.importer.load(node.modname, 1)

        #if module is None:
        #    print "Warning:", node.modname, "not imported."

        for name, alias in node.names:
            if name != "*":
                if module is not None and module.namespace.has_key(name):
                    attr = module[name]
                    self.store(alias or name, attr.value)
                    if isinstance(attr, Module) and not attr.loaded:
                        self.importer.load(attr.name)

                # Support the import of names from missing modules.

                else:
                    self.store(alias or name, UnresolvedName(name, node.modname, self))
            else:
                if module is not None:
                    for n in module.namespace.keys():
                        attr = module[n]
                        self.store(n, attr.value)
                        if isinstance(attr, Module) and not attr.loaded:
                            self.importer.load(attr.name)

        return None

    def visitFunction(self, node):
        function = Function(
            node.name,
            self.get_parent().full_name(),
            node.argnames,
            (node.flags & 4 != 0),
            (node.flags & 8 != 0),
            self,
            node
            )

        # Make a back reference from the node for code generation.

        node.unit = function

        self.namespaces.append(function)

        # Current namespace is the function.
        # Previous namespace is the class.

        if node.name == "__init__" and isinstance(self.namespaces[-2], Class):
            self.in_init = 1

        self.dispatch(node.code)
        self.in_init = 0
        self.namespaces.pop()

        self.store(node.name, function)
        return None

    visitGenExpr = NOP

    visitGenExprFor = NOP

    visitGenExprIf = NOP

    visitGenExprInner = NOP

    def visitGetattr(self, node):
        expr = self.dispatch(node.expr)
        if isinstance(expr, Attr):
            value = expr.value
            if isinstance(value, Module):
                return value.namespace.get(node.attrname)
            elif isinstance(value, UnresolvedName):
                return UnresolvedName(node.attrname, value.full_name(), self)
        return builtins.get(node.attrname)

    def visitGlobal(self, node):
        if self.namespaces:
            for name in node.names:
                self.namespaces[-1].make_global(name)

                # Record a global entry for the name in the module.

                if not self.has_key(name):
                    self[name] = Global()

    def visitIf(self, node):
        for test, body in node.tests:
            self.dispatch(body)
        if node.else_ is not None:
            self.dispatch(node.else_)
        return None

    visitIfExp = NOP

    def visitImport(self, node):
        if self.importer is None:
            raise InspectError(self.full_name(), node,
                "Please use the micropython.Importer class for code which uses the 'import' statement.")

        for name, alias in node.names:
            if alias is not None:
                self.store(alias, self.importer.load(name, 1) or UnresolvedName(None, name, self))
            else:
                self.store(name.split(".")[0], self.importer.load(name) or UnresolvedName(None, name.split(".")[0], self))

        return None

    visitInvert = NOP

    visitKeyword = NOP

    visitLambda = NOP

    visitLeftShift = NOP

    visitList = NOP

    visitListComp = NOP

    visitListCompFor = NOP

    visitListCompIf = NOP

    visitMod = NOP

    def visitModule(self, node):
        return self.dispatch(node.node)

    visitMul = NOP

    def visitName(self, node):
        name = node.name
        if self.namespaces and self.namespaces[-1].has_key(name):
            return self.namespaces[-1][name]
        elif self.has_key(name):
            return self[name]
        elif builtins.has_key(name):
            return builtins[name]
        else:
            return None

    visitNot = NOP

    visitOr = NOP

    visitPass = NOP

    visitPower = NOP

    visitPrint = NOP

    visitPrintnl = NOP

    visitRaise = NOP

    visitReturn = NOP

    visitRightShift = NOP

    visitSlice = NOP

    visitSliceobj = NOP

    def visitStmt(self, node):
        for n in node.nodes:
            self.dispatch(n)
        return None

    visitSub = NOP

    visitSubscript = NOP

    def visitTryExcept(self, node):
        self.dispatch(node.body)
        for name, var, n in node.handlers:
            self.dispatch(n)
        if node.else_ is not None:
            self.dispatch(node.else_)
        return None

    visitTryFinally = NOP

    visitTuple = NOP

    visitUnaryAdd = NOP

    visitUnarySub = NOP

    def visitWhile(self, node):
        self.in_loop = 1
        self.NOP(node)
        self.in_loop = 0

    visitWith = NOP

    visitYield = NOP

class Global:

    """
    A reference to an object assigned to a global from outside the module
    top-level.
    """

    pass

# Built-in types initialisation.

class Builtins(Module):

    "The special built-in types module."

    def __init__(self):
        Module.__init__(self, "__builtins__")
        self.loaded = 1
        self.module = None

        for key in ['ArithmeticError', 'AssertionError', 'AttributeError',
            'BaseException', 'DeprecationWarning', 'EOFError', 'Ellipsis',
            'EnvironmentError', 'Exception', 'False', 'FloatingPointError',
            'FutureWarning', 'GeneratorExit', 'IOError', 'ImportError',
            'ImportWarning', 'IndentationError', 'IndexError', 'KeyError',
            'KeyboardInterrupt', 'LookupError', 'MemoryError', 'NameError',
            'None', 'NotImplemented', 'NotImplementedError', 'OSError',
            'OverflowError', 'PendingDeprecationWarning', 'ReferenceError',
            'RuntimeError', 'RuntimeWarning', 'StandardError', 'StopIteration',
            'SyntaxError', 'SyntaxWarning', 'SystemError', 'SystemExit',
            'TabError', 'True', 'TypeError', 'UnboundLocalError',
            'UnicodeDecodeError', 'UnicodeEncodeError', 'UnicodeError',
            'UnicodeTranslateError', 'UnicodeWarning', 'UserWarning',
            'ValueError', 'Warning', 'ZeroDivisionError',
            'basestring', 'bool', 'buffer', 'complex', 'dict', 'file', 'float',
            'frozenset', 'int', # 'list',
            'long', 'object', 'set', 'slice', 'str',
            'tuple', 'type', 'unicode', 'xrange']:
            self.store(key, Class(key, self.full_name(), self))

        # NOTE: Temporary measure - provide detailed built-ins.

        cls = Class("list", self.full_name(), self)
        cls.set("__iter__", Function("__iter__", cls.full_name(), [], 0, 0, cls))
        self.store("list", cls)

        cls = Class("listiterator", self.full_name(), self)
        cls.set("next", Function("next", cls.full_name(), [], 0, 0, cls))
        self.store("listiterator", cls)

        # NOTE: Incomplete: some functions have more than one parameter.

        for key in ['__import__', 'abs', 'all', 'any', 'callable', 'chr',
            'classmethod', 'cmp', 'compile', 'delattr', 'dir', 'divmod',
            'enumerate', 'eval', 'execfile', 'filter', 'getattr', 'globals',
            'hasattr', 'hash', 'help', 'hex', 'id', 'input', 'isinstance',
            'issubclass', 'iter', 'len', 'locals', 'map', 'max', 'min', 'oct',
            'open', 'ord', 'pow', 'property', 'range', 'raw_input', 'reduce',
            'reload', 'repr', 'reversed', 'round', 'setattr', 'sorted',
            'staticmethod', 'sum', 'super', 'unichr', 'vars', 'zip']:
            self[key] = Function(key, self.full_name(), ['arg'], 0, 0, self)

    def store(self, name, obj):
        self.set(name, obj)
        self.all_objects.add(obj)

    def vacuum(self):
        pass

builtins = Builtins()

# vim: tabstop=4 expandtab shiftwidth=4
